from .tresnet import tresnet_m, tresnet_l, tresnet_xl, tresnet_m_448, tresnet_l_448, tresnet_xl_448

__all__ = ['tresnet_m', 'tresnet_l', 'tresnet_xl', 'tresnet_m_448', 'tresnet_l_448', 'tresnet_xl_448']

__version__ = '1.0.7'
__authors__ = 'ZHANG Zack'
